/* IBM_PROLOG_BEGIN_TAG                                                   */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* Program 5724-K69                                                       */
/*                                                                        */
/* COPYRIGHT International Business Machines Corp. 2004,2009              */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with                */
/* IBM Corp.                                                              */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */

// ABOUT:
// This overload provides a means to alter the runtime behaviour of 
// the gpu programs used by Google's "Google Earth" to render atmospheric haze
// Without this overload, multiple planets will be shown on a SVN rendering
// cluster. Alternatively selecting View->Atmosphere in Google Earth will 
// also ensure correct rendering. 
// This overload affects Server side rendering only

// DISCLAIMER:
// This software is provided without warranty, as is, for demonstration purposes only.
// Google Earth is a trademark of Google Inc. 

#if defined(WIN32)
// Dispose of annoying deprecation messages for strcpy etc.
#define _CRT_SECURE_NO_DEPRECATE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// wire.h contains the SVN OpenGL enumerants (e.g. GLCOLOR3F etc.) and must be included
#include <wire.h>

// Overload Variables 
// These hold the lengths of the shader programs, used for identification purposes
GLsizei firstGEarthProgramLength;
GLsizei secondGEarthProgramLength;
GLsizei thirdGEarthProgramLength;

// These hold the current IDs of the shader programs used by Google Earth
GLint firstGEarthProgram;
GLint secondGEarthProgram;
GLint thirdGEarthProgram;

// Using three functions
static void(*s_glProgramStringARB)(GLenum target, GLenum format, GLsizei len, const GLvoid *string);
static void(*s_glProgramLocalParameter4fvARB)(GLenum target, GLuint index, const GLfloat *params);
static void(*s_glGetProgramivARB)(GLenum target, GLenum param, GLint *value);
static void(*s_glGetFloatv)(GLenum pname, GLfloat *params);

// Overloading two functions
static void o_glProgramStringARB(GLenum target, GLenum format, GLsizei len, const GLvoid *string);
static void o_glProgramLocalParameter4fvARB(GLenum target, GLuint index, const GLfloat *params);

//
// Overload Function - replaces (or chains) sepcified OpenGL Functions
// to provide alternative functionality
//
void Overload(glOp *glOperationsTable, glOp *glSystemOperationsTable)
{
    FILE *fp = NULL;
    char *filename = NULL;
    char defaultFilename[] = "C:\\Program Files\\IBM\\IBM Deep Computing Visualisation\\examples\\overload\\googleearth\\googleearth_shaders.cfg";

    // Acquire a function pointer to access the DVGetDisplayName() SVN function
    char*(*_getDisplayName)() = 
        (char*(*)()) glOperationsTable[DVGETDISPLAYNAME];

    char* myDisplayName = _getDisplayName();

    // For Server Nodes myDisplayName will be the system's hostname as specified 
    // in the wall configuration file.
    // For the Client Application Node, myDisplayName will be "client".
    // If an error occurred myDisplayName will be NULL.
    if (myDisplayName == NULL)
    {
        fprintf(stderr, "Failed to obtain SVN system information. Overload aborted!\n");
        return;
    }
    else if ( strcmp(myDisplayName, "client") == 0 )
    {
        // Do not overload the clientside
        return;
    }
    
    filename = getenv("SVN_GECONFIG_FILE");
    if(!filename)
    {
        filename = defaultFilename;
    }    
    fp = fopen(filename,"r");
    if(!fp)
    {
#ifdef WIN32
        firstGEarthProgramLength    = 5028; // atmosphere_ground.arbvp1
        secondGEarthProgramLength   = 5038; // atmosphere_sky_model_space.arbvp1
        thirdGEarthProgramLength    = 4169; // atmosphere_sky_view_space.arbvp1
#else
        firstGEarthProgramLength    = 4864; // atmosphere_ground.arbvp1
        secondGEarthProgramLength   = 4867; // atmosphere_sky_model_space.arbvp1
        thirdGEarthProgramLength    = 4028; // atmosphere_sky_view_space.arbvp1
#endif
        fprintf(stderr, "Failed to load googleearth_shaders.cfg file, attempting defaults.\n");
    }
    else
    {
        char buffer[256];
        while(!feof(fp))
        {
            fgets(buffer, 256, fp);
            switch(buffer[0])
            {
            case 'f':
                sscanf(buffer, "firstlength %u\n", &firstGEarthProgramLength);
                break;
            case 's':
                sscanf(buffer, "secondlength %u\n", &secondGEarthProgramLength);
                break;
            case 't':
                sscanf(buffer, "thirdlength %u\n", &thirdGEarthProgramLength);
                break;
            default:
                break;
            }
        }
        fprintf(stderr, "Loaded googleearth_shaders.cfg file. L1 = %u, L2 = %u, L3 = %u\n", firstGEarthProgramLength, secondGEarthProgramLength, thirdGEarthProgramLength);
    }

    // Acquire a pointer to the 'true' OpenGL function for future use.
    // glOperationsTable are 'true' functions with SVN internal alterations
    // glSystemOperationsTable are the functions direct from the OpenGL driver
    fprintf(stderr, "SVN server rendering on display %s using \"Google Earth(tm) Shaders\" Overload\n", myDisplayName);

    s_glProgramStringARB = (void(*)(GLenum target, GLenum format, GLsizei len, const GLvoid *string)) glOperationsTable[GLPROGRAMSTRINGARB];
    s_glProgramLocalParameter4fvARB = (void(*)(GLenum target, GLuint index, const GLfloat *params)) glOperationsTable[GLPROGRAMLOCALPARAMETER4FVARB];
    s_glGetProgramivARB = (void(*)(GLenum target, GLenum param, GLint *value)) glOperationsTable[GLGETPROGRAMIVARB];
    s_glGetFloatv = (void(*)(GLenum pname, GLfloat *params)) glOperationsTable[GLGETFLOATV];
    
    // Now Overload the functions with the replacement versions
    glOperationsTable[GLPROGRAMSTRINGARB] = (glOp)o_glProgramStringARB;
    glOperationsTable[GLPROGRAMLOCALPARAMETER4FVARB] = (glOp)o_glProgramLocalParameter4fvARB;
}

//
// In glProgramString we identify the appropriate shaders by their length
// (stored in the config file and subject to change in the future)
//
static void
o_glProgramStringARB(GLenum target, GLenum format, GLsizei len, const GLvoid *string)
{
    s_glProgramStringARB(target, format, len, string);

    if(len == firstGEarthProgramLength) // 5028 / 4864
    {
        s_glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_BINDING_ARB, &firstGEarthProgram);
    }
    else if(len == secondGEarthProgramLength) // 5038 / 4867
    {
        s_glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_BINDING_ARB, &secondGEarthProgram);
    }
    else if(len == thirdGEarthProgramLength) // 4169 / 4028
    {
        s_glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_BINDING_ARB, &thirdGEarthProgram);
    }
}

//
// In glProgramLocalParameter4fvARB we must change the Modelview Matrix (passed as a param)
// to match the current wall scaling. Googleearth performs all matrix calcs in software and
// only uploads the final value - which we must then intercept and change.
//
static void
o_glProgramLocalParameter4fvARB(GLenum target, GLuint index, const GLfloat *params)
{
    GLfloat temp_buf[4];
    temp_buf[0] = params[0];
    temp_buf[1] = params[1];
    temp_buf[2] = params[2];
    temp_buf[3] = params[3];

    // Google perform ALL transforms in software and only
    // use glLoadMatrixf to set them for OGL (no tranlsates/rotates etc.)
    //
    // This is fine for most items, but the globe is drawn in a Vertex Program
    // The modelview/projection matrices are passed in as parameters for the program
    // We must modify the values to the appropriate wall-tiled transformed value
    //
    // We must track particular program ids (see glProgramStringARB)
    // (Windows deletes and creates new ones whenever we disable / enable atmosphere
    if(target == GL_VERTEX_PROGRAM_ARB)
    {
        int currentBoundProgram = 0;
        s_glGetProgramivARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_BINDING_ARB, &currentBoundProgram);

        if(currentBoundProgram == firstGEarthProgram || currentBoundProgram == secondGEarthProgram  ) // Windows
        {
            // the Modelviewprojection matrix is passed as a parameter (apply tiling transform
            // by replacing with the actual local modelview projection)
            if(index >=5 && index <= 8) 
            {
                GLfloat* mod = (GLfloat*)temp_buf;
                GLfloat model[16];
                GLfloat proj[16];

                s_glGetFloatv(GL_MODELVIEW_MATRIX, model);
                s_glGetFloatv(GL_PROJECTION_MATRIX, proj);

                switch(index)
                {            
                case 5: 
                    mod[0] = model[0]* proj[0] + model[1]* proj[4] + model[2]* proj[8] + model[3] *proj[12];
                    mod[1] = model[4]* proj[0] + model[5]* proj[4] + model[6]* proj[8] + model[7] *proj[12];
                    mod[2] = model[8]* proj[0] + model[9]* proj[4] + model[10]*proj[8] + model[11]*proj[12];
                    mod[3] = model[12]*proj[0] + model[13]*proj[4] + model[14]*proj[8] + model[15]*proj[12];
                    break;
                case 6: 
                    mod[0] = model[0]*proj[1]  + model[1]*proj[5]  + model[2]*proj[9]  + model[3]*proj[13];
                    mod[1] = model[4]*proj[1]  + model[5]*proj[5]  + model[6]*proj[9]  + model[7]*proj[13];
                    mod[2] = model[8]*proj[1]  + model[9]*proj[5]  + model[10]*proj[9]  + model[11]*proj[13];
                    mod[3] = model[12]*proj[1] + model[13]*proj[5] + model[14]*proj[9]  + model[15]*proj[13];
                    break;
                case 7: 
                    mod[0] = model[0]*proj[2] + model[1]*proj[6] + model[2]*proj[10] + model[3]*proj[14];
                    mod[1] = model[4]*proj[2] + model[5]*proj[6] + model[6]*proj[10] + model[7]*proj[14];
                    mod[2] = model[8]*proj[2] + model[9]*proj[6] + model[10]*proj[10] + model[11]*proj[14];
                    mod[3] = model[12]*proj[2] + model[13]*proj[6] + model[14]*proj[10] + model[15]*proj[14];
                    break;
                case 8: 
                    mod[0] = model[0]*proj[3] + model[1]*proj[7] + model[2]*proj[11] + model[3]*proj[15];
                    mod[1] = model[4]*proj[3] + model[5]*proj[7] + model[6]*proj[11] + model[7]*proj[15];
                    mod[2] = model[8]*proj[3] + model[9]*proj[7] + model[10]*proj[11] + model[11]*proj[15];
                    mod[3] = model[12]*proj[3] + model[13]*proj[7] + model[14]*proj[11] + model[15]*proj[15];
                    break;
                }
            }
        }
        else if(currentBoundProgram == thirdGEarthProgram)
        {
            // Just the projection matrix here
            if(index >=1 && index <= 4)
            {
                GLfloat* mod = (GLfloat*)temp_buf;
                GLfloat proj[16];
                s_glGetFloatv(GL_PROJECTION_MATRIX, proj);
                switch(index)
                {            
                case 1: 
                    mod[0] = proj[0]; mod[1] = proj[4]; mod[2] = proj[8]; mod[3] = proj[12];
                    break;
                case 2:
                    mod[0] = proj[1]; mod[1] = proj[5]; mod[2] = proj[9]; mod[3] = proj[13];
                    break;
                case 3: 
                    mod[0] = proj[2]; mod[1] = proj[6]; mod[2] = proj[10]; mod[3] = proj[14];
                    break;
                case 4: 
                    mod[0] = proj[3]; mod[1] = proj[7]; mod[2] = proj[11]; mod[3] = proj[15];
                    break;
                }
            }
        }
    }
    // Now call the REAL function (with potentially modified parameters)
    s_glProgramLocalParameter4fvARB(target, index, temp_buf);
}

