/* IBM_PROLOG_BEGIN_TAG                                                   */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* Program 5724-K69                                                       */
/*                                                                        */
/* COPYRIGHT International Business Machines Corp. 2004,2009              */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with                */
/* IBM Corp.                                                              */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */

// ABOUT:
// This overload demonstrates applying the Overload feature to only
// the rendering server nodes and not to the client application node

#include <stdio.h>
#include <string.h>
// wire.h contains the SVN OpenGL enumerants (e.g. GLCOLOR3F etc.) and must be included
#include "wire.h"

static void(*s_glColor3f)(float,float,float);
static void   o_glColor3f(float,float,float);


void Overload(glOp *glOperationsTable, glOp *glSystemOperationsTable)
{
    // Acquire a function pointer to access the DVGetDisplayName() SVN function
    char*(*_getDisplayName)() = 
        (char*(*)()) glOperationsTable[DVGETDISPLAYNAME];

    char* myDisplayName = _getDisplayName();

    // For Server Nodes myDisplayName will be the system's hostname as specified 
    // in the wall configuration file.
    // For the Client Application Node, myDisplayName will be "client".
    // If an error occurred myDisplayName will be NULL.

    if (myDisplayName==NULL)
    {
        fprintf(stderr,"Failed to obtain SVN system info;\n");
        return;
    }
    else if ( !strcmp(myDisplayName,"client") )
    {
        fprintf(stderr,"SVN client: aborting color inversion overload...\n");
        return;
    }

    fprintf(stderr,"SVN server rendering on display %s using color inversion overload... \n",myDisplayName);

    // Acquire a pointer to the 'true' OpenGL function for future use.
    s_glColor3f = (void(*)(float,float,float)) glOperationsTable [GLCOLOR3F];

    // Overload the set of OpenGL functions seen by the application with.
    // our custom glColor3f call instead
    glOperationsTable[GLCOLOR3F] = (glOp)o_glColor3f;
}

static void
o_glColor3f(float original_r, float original_g, float original_b)
{
    // Perform the inversion of the color components.
    float modified_r = 1 - original_r;
    float modified_g = 1 - original_g;
    float modified_b = 1 - original_b;

    // Call the 'true' glColor3f OpenGL function with modified values.
    (*s_glColor3f)(modified_r,modified_g,modified_b);
}
